/*
 * Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and
 * international Copyright laws.  Users and possessors of this source code
 * are hereby granted a nonexclusive, royalty-free license to use this code
 * in individual and commercial software.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION,  ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOURCE CODE.
 *
 * U.S. Government End Users.   This source code is a "commercial item" as
 * that term is defined at  48 C.F.R. 2.101 (OCT 1995), consisting  of
 * "commercial computer  software"  and "commercial computer software
 * documentation" as such terms are  used in 48 C.F.R. 12.212 (SEPT 1995)
 * and is provided to the U.S. Government only as a commercial end item.
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
 * source code with only those rights set forth herein.
 *
 * Any use of this source code in individual and commercial software must
 * include, in the user documentation and internal comments to the code,
 * the above Disclaimer and U.S. Government End Users Notice.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h> // for FLT_MAX

#include <cutil.h>
#include <cutil_math.h>

#include "CudaMath.h"
#include "dds.h"
#include "permutations.h"

// Definitions
//#define IFRAME "colors.ppm"
//#define IFRAME "lena-orig.ppm"
#define IFRAME "lena_std.ppm"
//#define IFRAME "lorikeet-orig.ppm"
//#define IFRAME "test16.ppm"

#define NUM_THREADS 64        // Number of threads per block.

#if __DEVICE_EMULATION__
#define __debugsync() __syncthreads()
#else
#define __debugsync()
#endif


template <class T> 
__device__ inline void swap(T & a, T & b)
{
    T tmp = a;
    a = b;
    b = tmp;
}


//__constant__ float3 kColorMetric = { 0.2126f, 0.7152f, 0.0722f };
__constant__ float3 kColorMetric = { 1.0f, 1.0f, 1.0f };


////////////////////////////////////////////////////////////////////////////////
// Sort colors
////////////////////////////////////////////////////////////////////////////////
__device__ void sortColors(const float * values, int * cmp)
{
	const int tid = threadIdx.x;

	cmp[tid] = (values[0] < values[tid]);
	cmp[tid] += (values[1] < values[tid]);
	cmp[tid] += (values[2] < values[tid]);
	cmp[tid] += (values[3] < values[tid]);
	cmp[tid] += (values[4] < values[tid]);
	cmp[tid] += (values[5] < values[tid]);
	cmp[tid] += (values[6] < values[tid]);
	cmp[tid] += (values[7] < values[tid]);
	cmp[tid] += (values[8] < values[tid]);
	cmp[tid] += (values[9] < values[tid]);
	cmp[tid] += (values[10] < values[tid]);
	cmp[tid] += (values[11] < values[tid]);
	cmp[tid] += (values[12] < values[tid]);
	cmp[tid] += (values[13] < values[tid]);
	cmp[tid] += (values[14] < values[tid]);
	cmp[tid] += (values[15] < values[tid]);
	
	// Resolve elements with the same index.
	if (tid > 0 && cmp[tid] == cmp[0]) ++cmp[tid];
	if (tid > 1 && cmp[tid] == cmp[1]) ++cmp[tid];
	if (tid > 2 && cmp[tid] == cmp[2]) ++cmp[tid];
	if (tid > 3 && cmp[tid] == cmp[3]) ++cmp[tid];
	if (tid > 4 && cmp[tid] == cmp[4]) ++cmp[tid];
	if (tid > 5 && cmp[tid] == cmp[5]) ++cmp[tid];
	if (tid > 6 && cmp[tid] == cmp[6]) ++cmp[tid];
	if (tid > 7 && cmp[tid] == cmp[7]) ++cmp[tid];
	if (tid > 8 && cmp[tid] == cmp[8]) ++cmp[tid];
	if (tid > 9 && cmp[tid] == cmp[9]) ++cmp[tid];
	if (tid > 10 && cmp[tid] == cmp[10]) ++cmp[tid];
	if (tid > 11 && cmp[tid] == cmp[11]) ++cmp[tid];
	if (tid > 12 && cmp[tid] == cmp[12]) ++cmp[tid];
	if (tid > 13 && cmp[tid] == cmp[13]) ++cmp[tid];
	if (tid > 14 && cmp[tid] == cmp[14]) ++cmp[tid];
}


////////////////////////////////////////////////////////////////////////////////
// Load color block to shared mem
////////////////////////////////////////////////////////////////////////////////
__device__ void loadColorBlock(const uint * image, float3 colors[16], float3 sums[16], int xrefs[16])
{
	const int bid = blockIdx.x;
	const int idx = threadIdx.x;

	__shared__ float dps[16];

	if (idx < 16)
	{
		// Read color and copy to shared mem.
		uint c = image[(bid) * 16 + idx];
	
		colors[idx].x = ((c >> 0) & 0xFF) * (1.0f / 255.0f);
		colors[idx].y = ((c >> 8) & 0xFF) * (1.0f / 255.0f);
		colors[idx].z = ((c >> 16) & 0xFF) * (1.0f / 255.0f);
		
		// No need to synchronize, 16 < warp size.
#if __DEVICE_EMULATION__
		} __debugsync(); if (idx < 16) {
#endif

        // Sort colors along the best fit line.
		colorSums(colors, sums);
		float3 axis = bestFitLine(colors, sums[0]);
        
        dps[idx] = dot(colors[idx], axis);
		
#if __DEVICE_EMULATION__
		} __debugsync(); if (idx < 16) {
#endif
		
		sortColors(dps, xrefs);
		
		float3 tmp = colors[idx];
		colors[xrefs[idx]] = tmp;
	}
}



////////////////////////////////////////////////////////////////////////////////
// Round color to RGB565 and expand
////////////////////////////////////////////////////////////////////////////////
inline __device__ float3 roundAndExpand(float3 v, ushort * w)
{
    v.x = rintf(__saturatef(v.x) * 31.0f);
    v.y = rintf(__saturatef(v.y) * 63.0f);
    v.z = rintf(__saturatef(v.z) * 31.0f);
    *w = ((ushort)v.x << 11) | ((ushort)v.y << 5) | (ushort)v.z;
    v.x *= 0.03227752766457f; // approximate integer bit expansion.
    v.y *= 0.01583151765563f;
    v.z *= 0.03227752766457f;
    return v;
}


__constant__ float alphaTable4[4] = { 9.0f, 0.0f, 6.0f, 3.0f };
__constant__ float alphaTable3[4] = { 4.0f, 0.0f, 2.0f, 2.0f };
__constant__ const int prods4[4] = { 0x090000,0x000900,0x040102,0x010402 };
__constant__ const int prods3[4] = { 0x040000,0x000400,0x040101,0x010401 };

////////////////////////////////////////////////////////////////////////////////
// Evaluate permutations
////////////////////////////////////////////////////////////////////////////////
static __device__ float evalPermutation4(const float3 * colors, uint permutation, ushort * start, ushort * end, float3 color_sum)
{
    // Compute endpoints using least squares.
//    float alpha2_sum = 0.0f;
//    float beta2_sum = 0.0f;
//    float alphabeta_sum = 0.0f;
    float3 alphax_sum = make_float3(0.0f, 0.0f, 0.0f);

	int akku = 0;

    // Compute alpha & beta for this permutation.
    for (int i = 0; i < 16; i++)
    {
		const uint bits = permutation >> (2*i);

		alphax_sum += alphaTable4[bits & 3] * colors[i];
	    akku += prods4[bits & 3];

//		float beta = (bits & 1);
//		if (bits & 2) beta = (1 + beta) * (1.0f / 3.0f);
//		float alpha = 1.0f - beta;

//		alpha2_sum += alpha * alpha;
//		beta2_sum += beta * beta;
//		alphabeta_sum += alpha * beta;
//		alphax_sum += alpha * colors[i];
    }

	float alpha2_sum = float(akku >> 16);
	float beta2_sum = float((akku >> 8) & 0xff);
	float alphabeta_sum = float((akku >> 0) & 0xff);
	float3 betax_sum = 9*color_sum - alphax_sum;


    // alpha2, beta2, alphabeta and factor could be precomputed for each permutation, but it's faster to recompute them.
    const float factor = 1.0f / (alpha2_sum * beta2_sum - alphabeta_sum * alphabeta_sum);

    float3 a = (alphax_sum * beta2_sum - betax_sum * alphabeta_sum) * factor;
    float3 b = (betax_sum * alpha2_sum - alphax_sum * alphabeta_sum) * factor;
    
    // Round a, b to the closest 5-6-5 color and expand...
    a = roundAndExpand(a, start);
    b = roundAndExpand(b, end);

    // compute the error
    float3 e = a * a * alpha2_sum + b * b * beta2_sum + 2.0f * (a * b * alphabeta_sum - a * alphax_sum - b * betax_sum);

    return (1.0f / 9.0f) * dot(e, kColorMetric);
}


static __device__ float evalPermutation3(const float3 * colors, uint permutation, ushort * start, ushort * end, float3 color_sum)
{
    // Compute endpoints using least squares.
//    float alpha2_sum = 0.0f;
//    float beta2_sum = 0.0f;
//    float alphabeta_sum = 0.0f;
    float3 alphax_sum = make_float3(0.0f, 0.0f, 0.0f);

	int akku = 0;

    // Compute alpha & beta for this permutation.
    for (int i = 0; i < 16; i++)
    {
        const uint bits = permutation >> (2*i);

		alphax_sum += alphaTable3[bits & 3] * colors[i];
	    akku += prods3[bits & 3];

//        float beta = (bits & 1);
//        if (bits & 2) beta = 0.5f;
//        float alpha = 1.0f - beta;
    
//        alpha2_sum += alpha * alpha;
//        beta2_sum += beta * beta;
//        alphabeta_sum += alpha * beta;
//        alphax_sum += alpha * colors[i];
    }

	float alpha2_sum = float(akku >> 16);
	float beta2_sum = float((akku >> 8) & 0xff);
	float alphabeta_sum = float((akku >> 0) & 0xff);
	float3 betax_sum = 4*color_sum - alphax_sum;

    const float factor = 1.0f / (alpha2_sum * beta2_sum - alphabeta_sum * alphabeta_sum);

    float3 a = (alphax_sum * beta2_sum - betax_sum * alphabeta_sum) * factor;
    float3 b = (betax_sum * alpha2_sum - alphax_sum * alphabeta_sum) * factor;
    
    // Round a, b to the closest 5-6-5 color and expand...
    a = roundAndExpand(a, start);
    b = roundAndExpand(b, end);

    // compute the error
    float3 e = a * a * alpha2_sum + b * b * beta2_sum + 2.0f * (a * b * alphabeta_sum - a * alphax_sum - b * betax_sum);

    return (1.0f / 4.0f) * dot(e, kColorMetric);
}



__device__ void evalAllPermutations(const float3 * colors, const uint * permutations, ushort & bestStart, ushort & bestEnd, uint & bestPermutation, float * errors, float3 color_sum)
{
	const int idx = threadIdx.x;
	
	float bestError = FLT_MAX;
	
	__shared__ uint s_permutations[160];

    for(int i = 0; i < 16; i++)
    {
		int pidx = idx + NUM_THREADS * i;
		if (pidx >= 992) break;
        
        ushort start, end;
        uint permutation = permutations[pidx];
        if (pidx < 160) s_permutations[pidx] = permutation;
                
        float error = evalPermutation4(colors, permutation, &start, &end, color_sum);
        
        if (error < bestError)
        {
            bestError = error;
            bestPermutation = permutation;
            bestStart = start;
            bestEnd = end;
        }
    }

    if (bestStart < bestEnd)
    {
        swap(bestEnd, bestStart);
        bestPermutation ^= 0x55555555;    // Flip indices.
    }

    for(int i = 0; i < 3; i++)
    {
		int pidx = idx + NUM_THREADS * i;
        if (pidx >= 160) break;
        
        ushort start, end;
        uint permutation = s_permutations[pidx];
        float error = evalPermutation3(colors, permutation, &start, &end, color_sum);
        
        if (error < bestError)
        {
            bestError = error;
            bestPermutation = permutation;
            bestStart = start;
            bestEnd = end;
            
            if (bestStart > bestEnd)
            {
                swap(bestEnd, bestStart);
                bestPermutation ^= (~bestPermutation >> 1) & 0x55555555;    // Flip indices.
            }
        }
    }

	errors[idx] = bestError;
}


////////////////////////////////////////////////////////////////////////////////
// Find index with minimum error
////////////////////////////////////////////////////////////////////////////////
__device__ int findMinError(float * errors)
{
	const int idx = threadIdx.x;

	__shared__ int indices[NUM_THREADS];
	indices[idx] = idx;

#if __DEVICE_EMULATION__
	for(int d = NUM_THREADS/2; d > 0; d >>= 1)
	{
		__syncthreads();
		
		if (idx < d)
		{
			float err0 = errors[idx];
			float err1 = errors[idx + d];
			
			if (err1 < err0) {
				errors[idx] = err1;
				indices[idx] = indices[idx + d];
			}
		}
	}

#else
	for(int d = NUM_THREADS/2; d > 32; d >>= 1)
	{
		__syncthreads();
		
		if (idx < d)
		{
			float err0 = errors[idx];
			float err1 = errors[idx + d];
			
			if (err1 < err0) {
				errors[idx] = err1;
				indices[idx] = indices[idx + d];
			}
		}
	}

	__syncthreads();

	// unroll last 6 iterations
	if (idx < 32)
	{
		if (errors[idx + 32] < errors[idx]) {
			errors[idx] = errors[idx + 32];
			indices[idx] = indices[idx + 32];
		}
		if (errors[idx + 16] < errors[idx]) {
			errors[idx] = errors[idx + 16];
			indices[idx] = indices[idx + 16];
		}
		if (errors[idx + 8] < errors[idx]) {
			errors[idx] = errors[idx + 8];
			indices[idx] = indices[idx + 8];
		}
		if (errors[idx + 4] < errors[idx]) {
			errors[idx] = errors[idx + 4];
			indices[idx] = indices[idx + 4];
		}
		if (errors[idx + 2] < errors[idx]) {
			errors[idx] = errors[idx + 2];
			indices[idx] = indices[idx + 2];
		}
		if (errors[idx + 1] < errors[idx]) {
			errors[idx] = errors[idx + 1];
			indices[idx] = indices[idx + 1];
		}
	}
#endif

	__syncthreads();

	return indices[0];
}


////////////////////////////////////////////////////////////////////////////////
// Save DXT block
////////////////////////////////////////////////////////////////////////////////
__device__ void saveBlockDXT1(ushort start, ushort end, uint permutation, int xrefs[16], uint2 * result)
{
	const int bid = blockIdx.x;

	if (start == end)
	{
		permutation = 0;
	}
	
	// Reorder permutation.
	uint indices = 0;
	for(int i = 0; i < 16; i++)
	{
		int ref = xrefs[i];
		indices |= ((permutation >> (2 * ref)) & 3) << (2 * i);
	}
	
	// Write endpoints.
	result[bid].x = (end << 16) | start;
	
	// Write palette indices.
	result[bid].y = indices;
}


////////////////////////////////////////////////////////////////////////////////
// Compress color block
////////////////////////////////////////////////////////////////////////////////
__global__ void compress(const uint * permutations, const uint * image, uint2 * result)
{
    const int idx = threadIdx.x;
    
	__shared__ float3 colors[16];
	__shared__ float3 sums[16];
	__shared__ int xrefs[16];
	
	loadColorBlock(image, colors, sums, xrefs);

    __syncthreads();

	ushort bestStart, bestEnd;
	uint bestPermutation;

	__shared__ float errors[NUM_THREADS];

	evalAllPermutations(colors, permutations, bestStart, bestEnd, bestPermutation, errors, sums[0]);

	// Use a parallel reduction to find minimum error.
	const int minIdx = findMinError(errors);
    
    __syncthreads();
    
    // Only write the result of the winner thread.
    if (idx == minIdx)
    {
		saveBlockDXT1(bestStart, bestEnd, bestPermutation, xrefs, result);
    }
}



////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char** argv) 
{
    CUT_DEVICE_INIT();

    // Load input image.
    unsigned char * data = NULL;
    uint w, h;

    char* image_path = cutFindFilePath(IFRAME, argv[0]);
    if (image_path == 0) {
        exit(EXIT_FAILURE);
    }

    if (!cutLoadPPM4ub( image_path, &data, &w, &h)) {
        printf("Error, unable to find source image\n");
        return 0;
    }

#if __DEVICE_EMULATION__
	// Reduce the image size so that it doesn't take so long on emulation.
	w = 4;
	h = 4;
#endif

    // Allocate input image.
    const uint memSize = w * h * 4;
    CUT_CONDITION( 0 != memSize );
    
    uint * image = (uint *) malloc(memSize);

    // Convert linear image to block linear. 
    for(int by = 0; by < h/4; by++) {
        for(int bx = 0; bx < w/4; bx++) {
            for (int i = 0; i < 16; i++) {
                const int x = i & 3;
                const int y = i / 4;
                image[(by * w/4 + bx) * 16 + i] = 
		    ((uint *)data)[(by * 4 + y) * 4 * (w/4) + bx * 4 + x];
            }
        }
    }

    // copy into global mem
    uint * d_data = NULL;
    CUDA_SAFE_CALL( cudaMalloc((void**) &d_data, memSize) );

    // Result
    uint * d_result = NULL;
    const uint compressedSize = ((w + 3) / 4) * ((h + 3) / 4) * 8;
    CUDA_SAFE_CALL( cudaMalloc((void**) &d_result, compressedSize) );
    uint * result = (uint *)malloc(compressedSize);

    // Compute permutations.
    uint permutations[1024];
    computePermutations(permutations);

    // Upload permutations.
    uint * d_permutations = NULL;    
    CUDA_SAFE_CALL( cudaMalloc((void**) &d_permutations, 1024 * sizeof(uint)) );
    CUDA_SAFE_CALL( cudaMemcpy(d_permutations, permutations, 1024 * sizeof(uint), 
			       cudaMemcpyHostToDevice) );


    uint timer;
    CUT_SAFE_CALL(cutCreateTimer(&timer));

    CUDA_SAFE_CALL(cudaThreadSynchronize()); 
    CUT_SAFE_CALL(cutStartTimer(timer));
    
    // Upload image.
    CUDA_SAFE_CALL( cudaMemcpy(d_data, image, memSize, cudaMemcpyHostToDevice) );

    uint blocks = ((w + 3) / 4) * ((h + 3) / 4);
    compress<<<blocks, NUM_THREADS>>>(d_permutations, d_data, (uint2 *)d_result);

    CUT_CHECK_ERROR("compress");

    CUDA_SAFE_CALL(cudaMemcpy(result, d_result, compressedSize, cudaMemcpyDeviceToHost));

	CUDA_SAFE_CALL(cudaThreadSynchronize());
    CUT_SAFE_CALL(cutStopTimer(timer));

    float time = cutGetTimerValue(timer);
    printf("Time %f msec\n", time);

	char output_filename[1024];
	strcpy(output_filename, image_path);
	strcpy(output_filename + strlen(image_path) - 3, "dds");

    // Write DDS file.
    FILE * fp = fopen(output_filename, "wb");

    DDSHeader header;
    header.fourcc = FOURCC_DDS;
    header.size = 124;
    header.flags  = (DDSD_WIDTH|DDSD_HEIGHT|DDSD_CAPS|DDSD_PIXELFORMAT|DDSD_LINEARSIZE);
    header.height = w;
    header.width = h;
    header.pitch = compressedSize;
    header.depth = 0;
    header.mipmapcount = 0;
    memset(header.reserved, 0, sizeof(header.reserved));
    header.pf.size = 32;
    header.pf.flags = DDPF_FOURCC;
    header.pf.fourcc = FOURCC_DXT1;
    header.pf.bitcount = 0;
    header.pf.rmask = 0;
    header.pf.gmask = 0;
    header.pf.bmask = 0;
    header.pf.amask = 0;
    header.caps.caps1 = DDSCAPS_TEXTURE;
    header.caps.caps2 = 0;
    header.caps.caps3 = 0;
    header.caps.caps4 = 0;
    header.notused = 0;

    fwrite(&header, sizeof(DDSHeader), 1, fp);
    fwrite(result, compressedSize, 1, fp);

    fclose(fp);

    // Free allocated memory.
    cutFree(image_path);
    cutFree(data);
    free(image);
    CUDA_SAFE_CALL(cudaFree(d_permutations));
    CUDA_SAFE_CALL(cudaFree(d_data));
    CUDA_SAFE_CALL(cudaFree(d_result));
    free(result);

    printf("Test PASSED\n");

    CUT_SAFE_CALL(cutDeleteTimer(timer));
    CUT_EXIT(argc, argv);
}

